@@ -3,13 +3,15 @@
 import logging
 import os
 import os.path
-import sys
 import configparser
 from typing import Dict, List, Tuple, Any, Optional, Union, Iterable
 from pathlib import Path
+from sqlfluff.core.plugin.host import get_plugin_manager
 
 import appdirs
 
+import toml
+
 # Instantiate the templater logger
 config_logger = logging.getLogger("sqlfluff.config")
 
@@ -68,7 +70,7 @@ def nested_combine(*dicts: dict) -> dict:
                     r[k] = nested_combine(r[k], d[k])
                 else:
                     raise ValueError(
-                        "Key {0!r} is a dict in one config but not another! PANIC: {1!r}".format(
+                        "Key {!r} is a dict in one config but not another! PANIC: {!r}".format(
                             k, d[k]
                         )
                     )
@@ -77,7 +79,7 @@ def nested_combine(*dicts: dict) -> dict:
     return r
 
 
-def dict_diff(left: dict, right: dict) -> dict:
+def dict_diff(left: dict, right: dict, ignore: Optional[List[str]] = None) -> dict:
     """Work out the difference between to dictionaries.
 
     Returns a dictionary which represents elements in the `left`
@@ -101,6 +103,8 @@ def dict_diff(left: dict, right: dict) -> dict:
     """
     buff: dict = {}
     for k in left:
+        if ignore and k in ignore:
+            continue
         # Is the key there at all?
         if k not in right:
             buff[k] = left[k]
@@ -109,8 +113,10 @@ def dict_diff(left: dict, right: dict) -> dict:
             continue
         # If it's not the same but both are dicts, then compare
         elif isinstance(left[k], dict) and isinstance(right[k], dict):
-            diff = dict_diff(left[k], right[k])
-            buff[k] = diff
+            diff = dict_diff(left[k], right[k], ignore=ignore)
+            # Only if the difference is not ignored it do we include it.
+            if diff:
+                buff[k] = diff
         # It's just different
         else:
             buff[k] = left[k]
@@ -139,6 +145,31 @@ class ConfigLoader:
             global_loader = cls()
         return global_loader
 
+    @classmethod
+    def _walk_toml(cls, config: Dict[str, Any], base_key=()):
+        """Recursively walk the nested config inside a TOML file."""
+        buff: List[tuple] = []
+        for k, v in config.items():
+            key = base_key + (k,)
+            if isinstance(v, dict):
+                buff.extend(cls._walk_toml(v, key))
+            else:
+                buff.append((key, v))
+
+        return buff
+
+    @classmethod
+    def _get_config_elems_from_toml(cls, fpath: str) -> List[Tuple[tuple, Any]]:
+        """Load a config from a TOML file and return a list of tuples.
+
+        The return value is a list of tuples, were each tuple has two elements,
+        the first is a tuple of paths, the second is the value at that path.
+        """
+        config = toml.load(fpath)
+        tool = config.get("tool", {}).get("sqlfluff", {})
+
+        return cls._walk_toml(tool)
+
     @staticmethod
     def _get_config_elems_from_file(fpath: str) -> List[Tuple[tuple, Any]]:
         """Load a config from a file and return a list of tuples.
@@ -151,7 +182,7 @@ class ConfigLoader:
             it reads config files.
 
         Note:
-            Any variable names ending with `_path`, will be attempted to be
+            Any variable names ending with `_path` or `_dir`, will be attempted to be
             resolved as relative paths to this config file. If that fails the
             string value will remain.
 
@@ -159,8 +190,7 @@ class ConfigLoader:
         buff: List[Tuple[tuple, Any]] = []
         # Disable interpolation so we can load macros
         kw: Dict = {}
-        if sys.version_info >= (3, 0):
-            kw["interpolation"] = None
+        kw["interpolation"] = None
         config = configparser.ConfigParser(**kw)
         # NB: We want to be case sensitive in how we read from files,
         # because jinja is also case sensitive. To do this we override
@@ -184,14 +214,13 @@ class ConfigLoader:
                 v = coerce_value(val)
 
                 # Attempt to resolve paths
-                if name.lower().endswith("_path"):
+                if name.lower().endswith(("_path", "_dir")):
                     # Try to resolve the path.
                     # Make the referenced path.
                     ref_path = os.path.join(os.path.dirname(fpath), val)
                     # Check if it exists, and if it does, replace the value with the path.
                     if os.path.exists(ref_path):
                         v = ref_path
-
                 # Add the name to the end of the key
                 buff.append((key + (name,), v))
         return buff
@@ -212,9 +241,7 @@ class ConfigLoader:
                     if isinstance(r[dp], dict):
                         r = r[dp]
                     else:
-                        raise ValueError(
-                            "Overriding config value with section! [{0}]".format(k)
-                        )
+                        raise ValueError(f"Overriding config value with section! [{k}]")
                 else:
                     r[dp] = {}
                     r = r[dp]
@@ -222,11 +249,12 @@ class ConfigLoader:
             r[n] = v
         return ctx
 
-    def load_default_config_file(self) -> dict:
+    def load_default_config_file(self, file_dir: str, file_name: str) -> dict:
         """Load the default config file."""
-        elems = self._get_config_elems_from_file(
-            os.path.join(os.path.dirname(__file__), "default_config.cfg")
-        )
+        if file_name == "pyproject.toml":
+            elems = self._get_config_elems_from_toml(os.path.join(file_dir, file_name))
+        else:
+            elems = self._get_config_elems_from_file(os.path.join(file_dir, file_name))
         return self._incorporate_vals({}, elems)
 
     def load_config_at_path(self, path: str) -> dict:
@@ -237,7 +265,13 @@ class ConfigLoader:
 
         # The potential filenames we would look for at this path.
         # NB: later in this list overwrites earlier
-        filename_options = ["setup.cfg", "tox.ini", "pep8.ini", ".sqlfluff"]
+        filename_options = [
+            "setup.cfg",
+            "tox.ini",
+            "pep8.ini",
+            ".sqlfluff",
+            "pyproject.toml",
+        ]
 
         configs: dict = {}
 
@@ -246,25 +280,45 @@ class ConfigLoader:
         else:
             p = os.path.dirname(path)
 
-        d = os.listdir(p)
+        d = os.listdir(os.path.expanduser(p))
         # iterate this way round to make sure things overwrite is the right direction
         for fname in filename_options:
             if fname in d:
-                elems = self._get_config_elems_from_file(os.path.join(p, fname))
+                if fname == "pyproject.toml":
+                    elems = self._get_config_elems_from_toml(os.path.join(p, fname))
+                else:
+                    elems = self._get_config_elems_from_file(os.path.join(p, fname))
                 configs = self._incorporate_vals(configs, elems)
 
         # Store in the cache
         self._config_cache[str(path)] = configs
         return configs
 
-    def load_user_appdir_config(self) -> dict:
-        """Load the config from the user's OS specific appdir config directory."""
+    @staticmethod
+    def _get_user_config_dir_path() -> str:
         appname = "sqlfluff"
         appauthor = "sqlfluff"
-        user_config_dir_path = appdirs.user_config_dir(appname, appauthor)
+
+        # On Mac OSX follow Linux XDG base dirs
+        # https://github.com/sqlfluff/sqlfluff/issues/889
+        user_config_dir_path = os.path.expanduser("~/.config/sqlfluff")
+        if appdirs.system == "darwin":
+            appdirs.system = "linux2"
+            user_config_dir_path = appdirs.user_config_dir(appname, appauthor)
+            appdirs.system = "darwin"
+
+        if not os.path.exists(user_config_dir_path):
+            user_config_dir_path = appdirs.user_config_dir(appname, appauthor)
+
+        return user_config_dir_path
+
+    def load_user_appdir_config(self) -> dict:
+        """Load the config from the user's OS specific appdir config directory."""
+        user_config_dir_path = self._get_user_config_dir_path()
         if os.path.exists(user_config_dir_path):
             return self.load_config_at_path(user_config_dir_path)
-        return {}
+        else:
+            return {}
 
     def load_user_config(self) -> dict:
         """Load the config from the user's home directory."""
@@ -338,7 +392,7 @@ class FluffConfig:
         self, configs: Optional[dict] = None, overrides: Optional[dict] = None
     ):
         self._overrides = overrides  # We only store this for child configs
-        defaults = ConfigLoader.get_global().load_default_config_file()
+        defaults = nested_combine(*get_plugin_manager().hook.load_default_config())
         self._configs = nested_combine(
             defaults, configs or {"core": {}}, {"core": overrides or {}}
         )
@@ -348,20 +402,24 @@ class FluffConfig:
         )
         # Deal with potential ignore parameters
         if self._configs["core"].get("ignore", None):
-            self._configs["core"]["ignore"] = self._configs["core"]["ignore"].split(",")
+            self._configs["core"]["ignore"] = self._split_comma_separated_string(
+                self._configs["core"]["ignore"]
+            )
         else:
             self._configs["core"]["ignore"] = []
         # Whitelists and blacklists
         if self._configs["core"].get("rules", None):
-            self._configs["core"]["rule_whitelist"] = self._configs["core"][
-                "rules"
-            ].split(",")
+            self._configs["core"][
+                "rule_whitelist"
+            ] = self._split_comma_separated_string(self._configs["core"]["rules"])
         else:
             self._configs["core"]["rule_whitelist"] = None
         if self._configs["core"].get("exclude_rules", None):
-            self._configs["core"]["rule_blacklist"] = self._configs["core"][
-                "exclude_rules"
-            ].split(",")
+            self._configs["core"][
+                "rule_blacklist"
+            ] = self._split_comma_separated_string(
+                self._configs["core"]["exclude_rules"]
+            )
         else:
             self._configs["core"]["rule_blacklist"] = None
         # Configure Recursion
@@ -443,7 +501,9 @@ class FluffConfig:
             or are different to the other.
 
         """
-        return dict_diff(self._configs, other._configs)
+        # We igonre some objects which are not meaningful in the comparison
+        # e.g. dialect_obj, which is generated on the fly.
+        return dict_diff(self._configs, other._configs, ignore=["dialect_obj"])
 
     def get(
         self, val: str, section: Union[str, Iterable[str]] = "core", default: Any = None
@@ -542,3 +602,15 @@ class FluffConfig:
         config_path = [elem.strip() for elem in config_line.split(":")]
         # Set the value
         self.set_value(config_path[:-1], config_path[-1])
+
+    def process_raw_file_for_config(self, raw_str: str):
+        """Process a full raw file for inline config and update self."""
+        # Scan the raw file for config commands.
+        for raw_line in raw_str.splitlines():
+            if raw_line.startswith("-- sqlfluff"):
+                # Found a in-file config command
+                self.process_inline_config(raw_line)
+
+    @staticmethod
+    def _split_comma_separated_string(raw_str: str) -> List[str]:
+        return [s.strip() for s in raw_str.split(",") if s.strip()]
