@@ -1,13 +1,14 @@
 """Base grammar, Ref, Anything and Nothing."""
 
 import copy
-from typing import List, NamedTuple, Optional, Union, Type, Tuple
+from dataclasses import dataclass
+from typing import List, Optional, Union, Type, Tuple, Any
 
 from sqlfluff.core.errors import SQLParseError
 from sqlfluff.core.string_helpers import curtail_string
 
-from sqlfluff.core.parser.segments import BaseSegment, EphemeralSegment
-from sqlfluff.core.parser.helpers import trim_non_code_segments
+from sqlfluff.core.parser.segments import BaseSegment, BracketedSegment, allow_ephemeral
+from sqlfluff.core.parser.helpers import trim_non_code_segments, iter_indices
 from sqlfluff.core.parser.match_result import MatchResult
 from sqlfluff.core.parser.match_logging import (
     parse_match_logging,
@@ -16,11 +17,32 @@ from sqlfluff.core.parser.match_logging import (
 from sqlfluff.core.parser.match_wrapper import match_wrapper
 from sqlfluff.core.parser.matchable import Matchable
 from sqlfluff.core.parser.context import ParseContext
+from sqlfluff.core.parser.parsers import StringParser
 
 # Either a Grammar or a Segment CLASS
 MatchableType = Union[Matchable, Type[BaseSegment]]
 
 
+@dataclass
+class BracketInfo:
+    """BracketInfo tuple for keeping track of brackets during matching.
+
+    This is used in BaseGrammar._bracket_sensitive_look_ahead_match but
+    defined here for type checking.
+    """
+
+    bracket: BaseSegment
+    segments: Tuple[BaseSegment, ...]
+
+    def to_segment(self, end_bracket):
+        """Turn the contained segments into a bracketed segment."""
+        return BracketedSegment(
+            segments=self.segments,
+            start_bracket=(self.bracket,),
+            end_bracket=end_bracket,
+        )
+
+
 def cached_method_for_parse_context(func):
     """A decorator to cache the output of this method for a given parse context.
 
@@ -66,6 +88,7 @@ class BaseGrammar(Matchable):
             # t: instance / f: class, ref, func
             (True, str, Ref.keyword),
             (True, BaseGrammar, lambda x: x),
+            (True, StringParser, lambda x: x),
             (False, BaseSegment, lambda x: x),
         ]
         # Get-out clause for None
@@ -78,7 +101,7 @@ class BaseGrammar(Matchable):
             ):
                 return init_func(elem)
         raise TypeError(
-            "Grammar element [{0!r}] was found of unexpected type [{1}] was found.".format(
+            "Grammar element [{!r}] was found of unexpected type [{}] was found.".format(
                 elem, type(elem)
             )
         )
@@ -119,26 +142,17 @@ class BaseGrammar(Matchable):
             for elem in args:
                 self._elements.append(self._resolve_ref(elem))
         else:
-            self._elements = args
+            self._elements = list(args)
 
         # Now we deal with the standard kwargs
         self.allow_gaps = allow_gaps
         self.optional = optional
-        self.ephemeral_segment = None
-        # Set up the ephemeral_segment if name is specified.
-        if ephemeral_name:
-            # Make the EphemeralSegment class. This is effectively syntactic sugar
-            # to allow us to avoid specifying a EphemeralSegment directly in a dialect.
-
-            # Copy self (*before* making the EphemeralSegment, but with everything else in place)
-            parse_grammar = copy.copy(self)
-            # Add the EphemeralSegment to self.
-            self.ephemeral_segment = EphemeralSegment.make(
-                match_grammar=None,
-                # Pass in the copy without the EphemeralSegment
-                parse_grammar=parse_grammar,
-                name=ephemeral_name,
-            )
+        # ephemeral_name is a flag to indicate whether we need to make an
+        # EphemeralSegment class. This is effectively syntactic sugar
+        # to allow us to avoid specifying a EphemeralSegment directly in a dialect.
+        # If this is the case, the actual segment construction happens in the
+        # match_wrapper.
+        self.ephemeral_name = ephemeral_name
 
     def is_optional(self):
         """Return whether this segment is optional.
@@ -148,6 +162,7 @@ class BaseGrammar(Matchable):
         return self.optional
 
     @match_wrapper()
+    @allow_ephemeral
     def match(self, segments: Tuple["BaseSegment", ...], parse_context: ParseContext):
         """Match a list of segments against this segment.
 
@@ -156,7 +171,7 @@ class BaseGrammar(Matchable):
         on the underlying class.
         """
         raise NotImplementedError(
-            "{0} has no match function implemented".format(self.__class__.__name__)
+            f"{self.__class__.__name__} has no match function implemented"
         )
 
     @cached_method_for_parse_context
@@ -218,6 +233,9 @@ class BaseGrammar(Matchable):
                 if res_match.matched_length > best_match_length:
                     best_match = res_match, matcher
                     best_match_length = res_match.matched_length
+            # We could stash segments here, but given we might have some successful
+            # matches here, we shouldn't, because they'll be mutated in the wrong way.
+            # Eventually there might be a performance gain from doing that sensibly here.
 
         # If we get here, then there wasn't a complete match. If we
         # has a best_match, return that.
@@ -290,20 +308,26 @@ class BaseGrammar(Matchable):
             # For existing compound segments, we should assume that within
             # that segment, things are internally consistent, that means
             # rather than enumerating all the individual segments of a longer
-            # one we just dump out the whole segment. This is a) faster and
+            # one we just dump out the whole segment, but splitting off the
+            # first element seperated by whitespace. This is a) faster and
             # also b) prevents some really horrible bugs with bracket matching.
             # See https://github.com/sqlfluff/sqlfluff/issues/433
-            str_buff = [seg.raw_upper for seg in segments]
+
+            def _trim_elem(seg):
+                s = seg.raw_upper.split(maxsplit=1)
+                return s[0] if s else ""
+
+            str_buff = [_trim_elem(seg) for seg in segments]
             match_queue = []
 
             for matcher, simple in simple_matchers:
                 # Simple will be a tuple of options
                 for simple_option in simple:
-                    try:
-                        buff_pos = str_buff.index(simple_option)
+                    # NOTE: We use iter_indices to make sure we capture
+                    # all instances of potential matches if there are many.
+                    # This is important for bracket counting.
+                    for buff_pos in iter_indices(str_buff, simple_option):
                         match_queue.append((matcher, buff_pos, simple_option))
-                    except ValueError:
-                        pass
 
             # Sort the match queue. First to process AT THE END.
             # That means we pop from the end.
@@ -421,22 +445,27 @@ class BaseGrammar(Matchable):
 
     @classmethod
     def _bracket_sensitive_look_ahead_match(
-        cls, segments, matchers, parse_context, start_bracket=None, end_bracket=None
-    ):
+        cls,
+        segments,
+        matchers,
+        parse_context,
+        start_bracket=None,
+        end_bracket=None,
+        bracket_pairs_set="bracket_pairs",
+    ) -> Tuple[Tuple[BaseSegment, ...], MatchResult, Optional[MatchableType]]:
         """Same as `_look_ahead_match` but with bracket counting.
 
         NB: Given we depend on `_look_ahead_match` we can also utilise
         the same performance optimisations which are implemented there.
 
+        bracket_pairs_set: Allows specific segments to override the available
+            bracket pairs. See the definition of "angle_bracket_pairs" in the
+            BigQuery dialect for additional context on why this exists.
+
         Returns:
             `tuple` of (unmatched_segments, match_object, matcher).
 
         """
-
-        class BracketInfo(NamedTuple):
-            bracket: BaseSegment
-            is_definite: bool
-
         # Type munging
         matchers = list(matchers)
         if isinstance(segments, BaseSegment):
@@ -450,30 +479,26 @@ class BaseGrammar(Matchable):
         # to the list of matchers. We get them from the relevant set on the
         # dialect. We use zip twice to "unzip" them. We ignore the first
         # argument because that's just the name.
-        _, start_bracket_refs, end_bracket_refs, definitely_bracket = zip(
-            *parse_context.dialect.sets("bracket_pairs")
+        _, start_bracket_refs, end_bracket_refs, persists = zip(
+            *parse_context.dialect.sets(bracket_pairs_set)
         )
-        # These are currently strings which need rehydrating
+        # These are matchables, probably StringParsers.
         start_brackets = [
             parse_context.dialect.ref(seg_ref) for seg_ref in start_bracket_refs
         ]
         end_brackets = [
             parse_context.dialect.ref(seg_ref) for seg_ref in end_bracket_refs
         ]
-        start_definite = list(definitely_bracket)
-        end_definite = list(definitely_bracket)
         # Add any bracket-like things passed as arguments
         if start_bracket:
             start_brackets += [start_bracket]
-            start_definite += [True]
         if end_bracket:
             end_brackets += [end_bracket]
-            end_definite += [True]
         bracket_matchers = start_brackets + end_brackets
 
         # Make some buffers
-        seg_buff = segments
-        pre_seg_buff = ()  # NB: Tuple
+        seg_buff: Tuple[BaseSegment, ...] = segments
+        pre_seg_buff: Tuple[BaseSegment, ...] = ()
         bracket_stack: List[BracketInfo] = []
 
         # Iterate
@@ -501,49 +526,78 @@ class BaseGrammar(Matchable):
                         # handle identical start and end brackets. For now, consider this
                         # a small, speculative investment in a possible future requirement.
                         if matcher in start_brackets and matcher not in end_brackets:
-                            # Same procedure as below in finding brackets.
+                            # Add any segments leading up to this to the previous bracket.
+                            bracket_stack[-1].segments += pre
+                            # Add a bracket to the stack and add the matches from the segment.
                             bracket_stack.append(
                                 BracketInfo(
                                     bracket=match.matched_segments[0],
-                                    is_definite=start_definite[
-                                        start_brackets.index(matcher)
-                                    ],
+                                    segments=match.matched_segments,
                                 )
                             )
-                            pre_seg_buff += pre
-                            pre_seg_buff += match.matched_segments
                             seg_buff = match.unmatched_segments
                             continue
                         elif matcher in end_brackets:
-                            # We've found an end bracket, remove it from the
-                            # stack and carry on.
-                            bracket_stack.pop()
-                            pre_seg_buff += pre
-                            pre_seg_buff += match.matched_segments
-                            seg_buff = match.unmatched_segments
-                            continue
+                            # Found an end bracket. Does its type match that of
+                            # the innermost start bracket? E.g. ")" matches "(",
+                            # "]" matches "[".
+                            # For the start bracket we don't have the matcher
+                            # but we can work out the name, so we use that for
+                            # the lookup.
+                            start_index = [
+                                bracket.name for bracket in start_brackets
+                            ].index(bracket_stack[-1].bracket.name)
+                            # For the end index, we can just look for the matcher
+                            end_index = end_brackets.index(matcher)
+                            bracket_types_match = start_index == end_index
+                            if bracket_types_match:
+                                # Yes, the types match. So we've found a
+                                # matching end bracket. Pop the stack, construct
+                                # a bracketed segment and carry
+                                # on.
+
+                                # Complete the bracketed info
+                                bracket_stack[-1].segments += (
+                                    pre + match.matched_segments
+                                )
+                                # Construct a bracketed segment (as a tuple) if allowed.
+                                persist_bracket = persists[end_brackets.index(matcher)]
+                                if persist_bracket:
+                                    new_segments: Tuple[BaseSegment, ...] = (
+                                        bracket_stack[-1].to_segment(
+                                            end_bracket=match.matched_segments
+                                        ),
+                                    )
+                                else:
+                                    new_segments = bracket_stack[-1].segments
+                                # Remove the bracket set from the stack
+                                bracket_stack.pop()
+                                # If we're still in a bracket, add the new segments to that bracket
+                                # Otherwise add them to the buffer
+                                if bracket_stack:
+                                    bracket_stack[-1].segments += new_segments
+                                else:
+                                    pre_seg_buff += new_segments
+                                seg_buff = match.unmatched_segments
+                                continue
+                            else:
+                                # The types don't match. Error.
+                                raise SQLParseError(
+                                    f"Found unexpected end bracket!, "
+                                    f"was expecting "
+                                    f"{end_brackets[start_index]}, "
+                                    f"but got {matcher}",
+                                    segment=match.matched_segments[0],
+                                )
+
                         else:
                             raise RuntimeError("I don't know how we get here?!")
                     else:
-                        # No match, we're in a bracket stack. Either this is an error,
-                        # OR we were mistaken in our initial identification of the opening
-                        # bracket. That's only allowed if `not definitely_bracket`.
-
-                        # Can we remove any brackets from the stack which aren't definites
-                        # to resolve the issue?
-                        for idx, elem in enumerate(reversed(bracket_stack)):
-                            if not elem.is_definite:
-                                del bracket_stack[-idx]
-                                # We don't change the string buffer, we assume that was ok.
-                                break
-                        else:
-                            # No we can't. We don't have a match and we're in a bracket stack.
-                            raise SQLParseError(
-                                "Couldn't find closing bracket for opening bracket.",
-                                segment=bracket_stack[-1].bracket,
-                            )
-                        # We have attempted a potential solution to the problem. Loop around.
-                        continue
+                        # No match, we're in a bracket stack. Error.
+                        raise SQLParseError(
+                            "Couldn't find closing bracket for opening bracket.",
+                            segment=bracket_stack[-1].bracket,
+                        )
                 else:
                     # No, we're open to more opening brackets or the thing(s)
                     # that we're otherwise looking for.
@@ -568,63 +622,59 @@ class BaseGrammar(Matchable):
                             bracket_stack.append(
                                 BracketInfo(
                                     bracket=match.matched_segments[0],
-                                    is_definite=start_definite[
-                                        start_brackets.index(matcher)
-                                    ],
+                                    segments=match.matched_segments,
                                 )
                             )
-                            # Add the matched elements and anything before it to the
-                            # pre segment buffer. Reset the working buffer.
+                            # The matched element has already been added to the bracket.
+                            # Add anything before it to the pre segment buffer.
+                            # Reset the working buffer.
                             pre_seg_buff += pre
-                            pre_seg_buff += match.matched_segments
                             seg_buff = match.unmatched_segments
                             continue
                         elif matcher in end_brackets:
-                            # each bracket with its "definite" attribute
-                            bracket_is_definite = end_definite[
-                                end_brackets.index(matcher)
-                            ]
-                            if bracket_is_definite:
-                                # We've found an unexpected end bracket!
-                                raise SQLParseError(
-                                    f"Found unexpected end bracket!, was expecting one of: {matchers}, but got {matcher}",
-                                    segment=match.matched_segments[0],
-                                )
-                            pre_seg_buff += pre
-                            pre_seg_buff += match.matched_segments
-                            seg_buff = match.unmatched_segments
-                            continue
+                            # We've found an unexpected end bracket! This is likely
+                            # because we're matching a section which should have ended.
+                            # If we had a match, it would have matched by now, so this
+                            # means no match.
+                            parse_match_logging(
+                                cls.__name__,
+                                "_bracket_sensitive_look_ahead_match",
+                                "UEXB",
+                                parse_context=parse_context,
+                                v_level=3,
+                                got=matcher,
+                            )
+                            # From here we'll drop out to the happy unmatched exit.
                         else:
                             # This shouldn't happen!?
                             raise NotImplementedError(
                                 "This shouldn't happen. Panic in _bracket_sensitive_look_ahead_match."
                             )
-                    else:
-                        # Not in a bracket stack, but no match. This is a happy
-                        # unmatched exit.
-                        return ((), MatchResult.from_unmatched(segments), None)
+                    # Not in a bracket stack, but no match.
+                    # From here we'll drop out to the happy unmatched exit.
             else:
                 # No we're at the end:
                 # Now check have we closed all our brackets?
                 if bracket_stack:
                     # No we haven't.
-                    # Check that the unclosed brackets are definite
-                    definite_bracket_stack = [b for b in bracket_stack if b.is_definite]
-                    if definite_bracket_stack:
-                        raise SQLParseError(
-                            f"Couldn't find closing bracket for opened brackets: `{bracket_stack}`.",
-                            segment=bracket_stack[-1].bracket,
-                        )
+                    raise SQLParseError(
+                        f"Couldn't find closing bracket for opened brackets: `{bracket_stack}`.",
+                        segment=bracket_stack[-1].bracket,
+                    )
 
-                # We at the end but without a bracket left open. This is a
-                # friendly unmatched return.
-                return ((), MatchResult.from_unmatched(segments), None)
+            # This is the happy unmatched path. This occurs when:
+            # - We reached the end with no open brackets.
+            # - No match while outside a bracket stack.
+            # - We found an unexpected end bracket before matching something interesting.
+            # We return with the mutated segments so we can
+            # reuse any bracket matching.
+            return ((), MatchResult.from_unmatched(pre_seg_buff + seg_buff), None)
 
     def __str__(self):
         return repr(self)
 
     def __repr__(self):
-        return "<{0}: [{1}]>".format(
+        return "<{}: [{}]>".format(
             self.__class__.__name__,
             curtail_string(
                 ", ".join(curtail_string(repr(elem), 40) for elem in self._elements),
@@ -632,6 +682,90 @@ class BaseGrammar(Matchable):
             ),
         )
 
+    def __eq__(self, other):
+        """Two grammars are equal if their elements and types are equal.
+
+        NOTE: This could potentially mean that two grammars with
+        the same elements but _different configuration_ will be
+        classed as the same. If this matters for your use case,
+        consider extending this function.
+
+        e.g. `OneOf(foo) == OneOf(foo, optional=True)`
+        """
+        return type(self) is type(other) and self._elements == other._elements
+
+    def copy(
+        self,
+        insert: Optional[list] = None,
+        at: Optional[int] = None,
+        before: Optional[Any] = None,
+        remove: Optional[list] = None,
+        **kwargs,
+    ):
+        """Create a copy of this grammar, optionally with differences.
+
+        This is mainly used in dialect inheritance.
+
+
+        Args:
+            insert (:obj:`list`, optional): Matchable elements to
+                insert. This is inserted pre-expansion so can include
+                unexpanded elements as normal.
+            at (:obj:`int`, optional): The position in the elements
+                to insert the item. Defaults to `None` which means
+                insert at the end of the elements.
+            before (optional): An alternative to _at_ to determine the
+                position of an insertion. Using this inserts the elements
+                immediately before the position of this element.
+                Note that this is not an _index_ but an element to look
+                for (i.e. a Segment or Grammar which will be compared
+                with other elements for equality).
+            remove (:obj:`list`, optional): A list of individual
+                elements to remove from a grammar. Removal is done
+                *after* insertion so that order is preserved.
+                Elements are searched for individually.
+
+        """
+        # Copy only the *grammar* elements. The rest comes through
+        # as is because they should just be classes rather than
+        # instances.
+        new_elems = [
+            elem.copy() if isinstance(elem, BaseGrammar) else elem
+            for elem in self._elements
+        ]
+        if insert:
+            if at is not None and before is not None:
+                raise ValueError(
+                    "Cannot specify `at` and `before` in BaseGrammar.copy()."
+                )
+            if before is not None:
+                try:
+                    idx = new_elems.index(before)
+                except ValueError:
+                    raise ValueError(
+                        "Could not insert {} in copy of {}. {} not Found.".format(
+                            insert, self, before
+                        )
+                    )
+                new_elems = new_elems[:idx] + insert + new_elems[idx:]
+            elif at is None:
+                new_elems = new_elems + insert
+            else:
+                new_elems = new_elems[:at] + insert + new_elems[at:]
+        if remove:
+            for elem in remove:
+                try:
+                    new_elems.remove(elem)
+                except ValueError:
+                    raise ValueError(
+                        "Could not remove {} from copy of {}. Not Found.".format(
+                            elem, self
+                        )
+                    )
+        new_seg = copy.copy(self)
+        new_seg._elements = new_elems
+        return new_seg
+
 
 class Ref(BaseGrammar):
     """A kind of meta-grammar that references other grammars by name at runtime."""
@@ -659,7 +793,7 @@ class Ref(BaseGrammar):
             return self._elements[0]
         else:
             raise ValueError(
-                "Ref grammar can only deal with precisely one element for now. Instead found {0!r}".format(
+                "Ref grammar can only deal with precisely one element for now. Instead found {!r}".format(
                     self._elements
                 )
             )
@@ -673,11 +807,12 @@ class Ref(BaseGrammar):
             raise ReferenceError("No Dialect has been provided to Ref grammar!")
 
     def __repr__(self):
-        return "<Ref: {0}{1}>".format(
+        return "<Ref: {}{}>".format(
             ", ".join(self._elements), " [opt]" if self.is_optional() else ""
         )
 
     @match_wrapper(v_level=4)  # Log less for Ref
+    @allow_ephemeral
     def match(self, segments, parse_context):
         """Match a list of segments against this segment.
 
@@ -691,13 +826,10 @@ class Ref(BaseGrammar):
         elem = self._get_elem(dialect=parse_context.dialect)
 
         if not elem:
-            raise ValueError(
-                "Null Element returned! _elements: {0!r}".format(self._elements)
-            )
+            raise ValueError(f"Null Element returned! _elements: {self._elements!r}")
 
         # First check against the efficiency Cache.
-        # We used to use seg_to_tuple here, but it was too slow,
-        # so instead we rely on segments not being mutated within a given
+        # We rely on segments not being mutated within a given
         # match cycle and so the ids should continue to refer to unchanged
         # objects.
         seg_tuple = (id(seg) for seg in segments)
