@@ -1,7 +1,7 @@
-"""Implements the base crawler which all the rules are based on.
+"""Implements the base rule class.
 
-Crawlers, crawl through the trees returned by the parser and
-evaluate particular rules.
+Rules crawl through the trees returned by the parser and evaluate particular
+rules.
 
 The intent is that it should be possible for the rules to be expressed
 as simply as possible, with as much of the complexity abstracted away.
@@ -16,11 +16,17 @@ missing.
 
 import copy
 import logging
+import pathlib
+import re
+from typing import Optional, List, Tuple, TYPE_CHECKING
 from collections import namedtuple
 
-from sqlfluff.core.parser import RawSegment, KeywordSegment, BaseSegment, SymbolSegment
+from sqlfluff.core.parser import BaseSegment
 from sqlfluff.core.errors import SQLLintError
 
+if TYPE_CHECKING:
+    from sqlfluff.core.templaters import TemplatedFile
+
 # The ghost of a rule (mostly used for testing)
 RuleGhost = namedtuple("RuleGhost", ["code", "description"])
 
@@ -33,11 +39,11 @@ class RuleLoggingAdapter(logging.LoggerAdapter):
 
     def process(self, msg, kwargs):
         """Add the code element to the logging message before emit."""
-        return "[%s] %s" % (self.extra["code"], msg), kwargs
+        return "[{}] {}".format(self.extra["code"], msg), kwargs
 
 
 class LintResult:
-    """A class to hold the results of a crawl operation.
+    """A class to hold the results of a rule evaluation.
 
     Args:
         anchor (:obj:`BaseSegment`, optional): A segment which represents
@@ -48,7 +54,7 @@ class LintResult:
             fixes which would correct this issue. If not present then it's
             assumed that this issue will have to manually fixed.
         memory (:obj:`dict`, optional): An object which stores any working
-            memory for the crawler. The `memory` returned in any `LintResult`
+            memory for the rule. The `memory` returned in any `LintResult`
             will be passed as an input to the next segment to be crawled.
         description (:obj:`str`, optional): A description of the problem
             identified as part of this result. This will override the
@@ -97,15 +103,17 @@ class LintFix:
             to be moved *after* the edit), for an `edit` it implies the segment
             to be replaced.
         edit (:obj:`BaseSegment`, optional): For `edit` and `create` fixes, this
-            hold the segment, or iterable of segments to create to replace at the
+            hold the segment, or iterable of segments to create or replace at the
             given `anchor` point.
 
     """
 
-    def __init__(self, edit_type, anchor, edit=None):
+    def __init__(self, edit_type, anchor: BaseSegment, edit=None):
         if edit_type not in ["create", "edit", "delete"]:
-            raise ValueError("Unexpected edit_type: {0}".format(edit_type))
+            raise ValueError(f"Unexpected edit_type: {edit_type}")
         self.edit_type = edit_type
+        if not anchor:
+            raise ValueError("Fixes must provide an anchor.")
         self.anchor = anchor
         # Coerce to list
         if isinstance(edit, BaseSegment):
@@ -117,9 +125,17 @@ class LintFix:
         # the parsed structure.
         self.edit = copy.deepcopy(edit)
         if self.edit:
+            # Check that any edits don't have a position marker set.
+            # We should rely on realignment to make position markers.
             # Strip position markers of anything enriched, otherwise things can get blurry
             for seg in self.edit:
-                seg.pos_marker = seg.pos_marker.strip()
+                if seg.pos_marker:
+                    # Developer warning.
+                    rules_logger.debug(
+                        "Developer Note: Edit segment found with preset position marker. "
+                        "These should be unset and calculated later."
+                    )
+                    seg.pos_marker = None
         # Once stripped, we shouldn't replace any markers because
         # later code may rely on them being accurate, which we
         # can't guarantee with edits.
@@ -145,7 +161,7 @@ class LintFix:
 
     def __repr__(self):
         if self.edit_type == "delete":
-            detail = "delete:{0!r}".format(self.anchor.raw)
+            detail = f"delete:{self.anchor.raw!r}"
         elif self.edit_type in ("edit", "create"):
             if hasattr(self.edit, "raw"):
                 new_detail = self.edit.raw
@@ -153,12 +169,12 @@ class LintFix:
                 new_detail = "".join(s.raw for s in self.edit)
 
             if self.edit_type == "edit":
-                detail = "edt:{0!r}->{1!r}".format(self.anchor.raw, new_detail)
+                detail = f"edt:{self.anchor.raw!r}->{new_detail!r}"
             else:
-                detail = "create:{0!r}".format(new_detail)
+                detail = f"create:{new_detail!r}"
         else:
             detail = ""
-        return "<LintFix: {0} @{1} {2}>".format(
+        return "<LintFix: {} @{} {}>".format(
             self.edit_type, self.anchor.pos_marker, detail
         )
 
@@ -178,8 +194,8 @@ class LintFix:
         return True
 
 
-class BaseCrawler:
-    """The base class for a crawler, of which all rules are derived from.
+class BaseRule:
+    """The base class for a rule.
 
     Args:
         code (:obj:`str`): The identifier for this rule, used in inclusion
@@ -190,11 +206,12 @@ class BaseCrawler:
     """
 
     _works_on_unparsable = True
+    targets_templated = False
 
     def __init__(self, code, description, **kwargs):
         self.description = description
         self.code = code
-        # kwargs represents the config passed to the crawler. Add all kwargs as class attributes
+        # kwargs represents the config passed to the rule. Add all kwargs as class attributes
         # so they can be accessed in rules which inherit from this class
         for key, value in kwargs.items():
             self.__dict__[key] = value
@@ -208,13 +225,13 @@ class BaseCrawler:
                 if keyword not in kwargs.keys():
                     raise ValueError(
                         (
-                            "Unrecognized config '{0}' for Rule {1}. If this "
+                            "Unrecognized config '{}' for Rule {}. If this "
                             "is a new option, please add it to "
                             "`default_config.cfg`"
                         ).format(keyword, code)
                     )
         except AttributeError:
-            self.logger.info("No config_keywords defined for {0}".format(code))
+            self.logger.info(f"No config_keywords defined for {code}")
 
     def _eval(self, **kwargs):
         """Evaluate this rule against the current context.
@@ -236,7 +253,7 @@ class BaseCrawler:
         """
         raise NotImplementedError(
             (
-                "{0} has not had its `eval` function defined. This is a problem "
+                "{} has not had its `eval` function defined. This is a problem "
                 "with the rule setup."
             ).format(self.__class__.__name__)
         )
@@ -250,6 +267,8 @@ class BaseCrawler:
         siblings_post=None,
         raw_stack=None,
         memory=None,
+        fname=None,
+        templated_file: Optional["TemplatedFile"] = None,
     ):
         """Recursively perform the crawl operation on a given segment.
 
@@ -259,7 +278,7 @@ class BaseCrawler:
         """
         # parent stack should be a tuple if it exists
 
-        # crawlers, should evaluate on segments FIRST, before evaluating on their
+        # Rules should evaluate on segments FIRST, before evaluating on their
         # children. They should also return a list of violations.
 
         parent_stack = parent_stack or ()
@@ -267,8 +286,8 @@ class BaseCrawler:
         siblings_post = siblings_post or ()
         siblings_pre = siblings_pre or ()
         memory = memory or {}
-        vs = []
-        fixes = []
+        vs: List[SQLLintError] = []
+        fixes: List[LintFix] = []
 
         # First, check whether we're looking at an unparsable and whether
         # this rule will still operate on that.
@@ -286,6 +305,8 @@ class BaseCrawler:
                 raw_stack=raw_stack,
                 memory=memory,
                 dialect=dialect,
+                path=pathlib.Path(fname) if fname else None,
+                templated_file=templated_file,
             )
         # Any exception at this point would halt the linter and
         # cause the user to get no results
@@ -293,6 +314,7 @@ class BaseCrawler:
             self.logger.critical(
                 f"Applying rule {self.code} threw an Exception: {e}", exc_info=True
             )
+            exception_line, _ = segment.pos_marker.source_position()
             vs.append(
                 SQLLintError(
                     rule=self,
@@ -302,13 +324,15 @@ class BaseCrawler:
                         f"""Unexpected exception: {str(e)};
                         Could you open an issue at https://github.com/sqlfluff/sqlfluff/issues ?
                         You can ignore this exception for now, by adding '--noqa: {self.code}' at the end
-                        of line {segment.pos_marker.line_no}
+                        of line {exception_line}
                         """
                     ),
                 )
             )
             return vs, raw_stack, fixes, memory
 
+        new_lerrs = []
+        new_fixes = []
         if res is None:
             # Assume this means no problems (also means no memory)
             pass
@@ -317,8 +341,8 @@ class BaseCrawler:
             memory = res.memory
             lerr = res.to_linting_error(rule=self)
             if lerr:
-                vs.append(lerr)
-            fixes += res.fixes
+                new_lerrs = [lerr]
+            new_fixes = res.fixes
         elif isinstance(res, list) and all(
             isinstance(elem, LintResult) for elem in res
         ):
@@ -328,15 +352,24 @@ class BaseCrawler:
             for elem in res:
                 lerr = elem.to_linting_error(rule=self)
                 if lerr:
-                    vs.append(lerr)
-                fixes += elem.fixes
+                    new_lerrs.append(lerr)
+                new_fixes += elem.fixes
         else:
             raise TypeError(
-                "Got unexpected result [{0!r}] back from linting rule: {1!r}".format(
+                "Got unexpected result [{!r}] back from linting rule: {!r}".format(
                     res, self.code
                 )
             )
 
+        for lerr in new_lerrs:
+            self.logger.debug("!! Violation Found: %r", lerr.description)
+        for fix in new_fixes:
+            self.logger.debug("!! Fix Proposed: %r", fix)
+
+        # Consume the new results
+        vs += new_lerrs
+        fixes += new_fixes
+
         # The raw stack only keeps track of the previous raw segments
         if len(segment.segments) == 0:
             raw_stack += (segment,)
@@ -352,6 +385,8 @@ class BaseCrawler:
                 raw_stack=raw_stack,
                 memory=memory,
                 dialect=dialect,
+                fname=fname,
+                templated_file=templated_file,
             )
             vs += dvs
             fixes += child_fixes
@@ -396,37 +431,14 @@ class BaseCrawler:
         # no subsegments to check. Return None.
         return None
 
-    @classmethod
-    def make_whitespace(cls, raw, pos_marker):
-        """Make a whitespace segment."""
-        WhitespaceSegment = RawSegment.make(" ", name="whitespace", type="whitespace")
-        return WhitespaceSegment(raw=raw, pos_marker=pos_marker)
-
-    @classmethod
-    def make_newline(cls, pos_marker, raw=None):
-        """Make a newline segment."""
-        # Default the newline to \n
-        raw = raw or "\n"
-        nls = RawSegment.make("\n", name="newline", type="newline")
-        return nls(raw=raw, pos_marker=pos_marker)
-
-    @classmethod
-    def make_keyword(cls, raw, pos_marker):
-        """Make a keyword segment."""
-        # For the name of the segment, we force the string to lowercase.
-        kws = KeywordSegment.make(raw.lower())
-        # At the moment we let the rule dictate *case* here.
-        return kws(raw=raw, pos_marker=pos_marker)
-
-    @classmethod
-    def make_symbol(cls, raw, pos_marker, seg_type, name=None):
-        """Make a symbol segment."""
-        # For the name of the segment, we force the string to lowercase.
-        symbol_seg = SymbolSegment.make(
-            raw.lower(), name=name or seg_type, type=seg_type
-        )
-        # At the moment we let the rule dictate *case* here.
-        return symbol_seg(raw=raw, pos_marker=pos_marker)
+    @staticmethod
+    def matches_target_tuples(seg: BaseSegment, target_tuples: List[Tuple[str, str]]):
+        """Does the given segment match any of the given type tuples."""
+        if seg.name in [elem[1] for elem in target_tuples if elem[0] == "name"]:
+            return True
+        elif seg.is_type(*[elem[1] for elem in target_tuples if elem[0] == "type"]):
+            return True
+        return False
 
 
 class RuleSet:
@@ -463,18 +475,21 @@ class RuleSet:
         Config options can also be checked for a specific rule e.g L010.
         """
         rule_config = config.get_section("rules")
-
         for config_name, info_dict in self.config_info.items():
             config_option = (
                 rule_config.get(config_name)
                 if not rule
                 else rule_config.get(rule).get(config_name)
             )
-            valid_options = info_dict["validation"]
-            if config_option not in valid_options and config_option is not None:
+            valid_options = info_dict.get("validation")
+            if (
+                valid_options
+                and config_option not in valid_options
+                and config_option is not None
+            ):
                 raise ValueError(
                     (
-                        "Invalid option '{0}' for {1} configuration. Must be one of {2}"
+                        "Invalid option '{}' for {} configuration. Must be one of {}"
                     ).format(
                         config_option,
                         config_name,
@@ -482,13 +497,27 @@ class RuleSet:
                     )
                 )
 
-    def register(self, cls):
+    @property
+    def valid_rule_name_regex(self):
+        """Defines the accepted pattern for rule names.
+
+        The first group captures the plugin name (optional), which
+        must be capitalized.
+        The second group captures the rule code.
+
+        Examples of valid rule names:
+        * Rule_PluginName_L001
+        * Rule_L001
+        """
+        return re.compile(r"Rule_?([A-Z]{1}[a-zA-Z]+)?_([A-Z][0-9]{3})")
+
+    def register(self, cls, plugin=None):
         """Decorate a class with this to add it to the ruleset.
 
         .. code-block:: python
 
            @myruleset.register
-           class Rule_L001(BaseCrawler):
+           class Rule_L001(BaseRule):
                "Description of rule."
 
                def eval(self, **kwargs):
@@ -502,23 +531,28 @@ class RuleSet:
         :exc:`ValueError`.
 
         """
-        elems = cls.__name__.split("_")
+        rule_name_match = self.valid_rule_name_regex.match(cls.__name__)
         # Validate the name
-        if len(elems) != 2 or elems[0] != "Rule" or len(elems[1]) != 4:
+        if not rule_name_match:
             raise ValueError(
                 (
-                    "Tried to register rule on set {0!r} with unexpected " "format: {1}"
+                    "Tried to register rule on set {!r} with unexpected "
+                    "format: {}, format should be: Rule_PluginName_L123 (for plugins) "
+                    "or Rule_L123 (for core rules)."
                 ).format(self.name, cls.__name__)
             )
 
-        code = elems[1]
+        plugin_name, code = rule_name_match.groups()
         # If the docstring is multiline, then we extract just summary.
         description = cls.__doc__.split("\n")[0]
 
+        if plugin_name:
+            code = f"{plugin_name}_{code}"
+
         # Keep track of the *class* in the register. Don't instantiate yet.
         if code in self._register:
             raise ValueError(
-                "Rule {0!r} has already been registered on RuleSet {1!r}!".format(
+                "Rule {!r} has already been registered on RuleSet {!r}!".format(
                     code, self.name
                 )
             )
@@ -527,14 +561,14 @@ class RuleSet:
         # Make sure we actually return the original class
         return cls
 
-    def get_rulelist(self, config):
+    def get_rulelist(self, config) -> List[BaseRule]:
         """Use the config to return the appropriate rules.
 
         We use the config both for whitelisting and blacklisting, but also
         for configuring the rules given the given config.
 
         Returns:
-            :obj:`list` of instantiated :obj:`BaseCrawler`.
+            :obj:`list` of instantiated :obj:`BaseRule`.
 
         """
         # Validate all generic rule configs
@@ -548,7 +582,7 @@ class RuleSet:
         ]
         if any(whitelisted_unknown_rule_codes):
             rules_logger.warning(
-                "Tried to whitelist unknown rules: {0!r}".format(
+                "Tried to whitelist unknown rules: {!r}".format(
                     whitelisted_unknown_rule_codes
                 )
             )
@@ -558,7 +592,7 @@ class RuleSet:
         ]
         if any(blacklisted_unknown_rule_codes):
             rules_logger.warning(
-                "Tried to blacklist unknown rules: {0!r}".format(
+                "Tried to blacklist unknown rules: {!r}".format(
                     blacklisted_unknown_rule_codes
                 )
             )
