@@ -2,6 +2,7 @@
 
 import os.path
 import logging
+import importlib.util
 from typing import Iterator, Tuple, Optional
 
 from jinja2.sandbox import SandboxedEnvironment
@@ -9,7 +10,6 @@ from jinja2 import meta, TemplateSyntaxError, TemplateError
 import jinja2.nodes
 
 from sqlfluff.core.errors import SQLTemplaterError
-from sqlfluff.core.parser import FilePositionMarker
 
 from sqlfluff.core.templaters.base import (
     register_templater,
@@ -56,12 +56,12 @@ class JinjaTemplater(PythonTemplater):
         """Take a path and extract macros from it."""
         # Does the path exist? It should as this check was done on config load.
         if not os.path.exists(path):
-            raise ValueError("Path does not exist: {0}".format(path))
+            raise ValueError(f"Path does not exist: {path}")
 
         macro_ctx = {}
         if os.path.isfile(path):
             # It's a file. Extract macros from it.
-            with open(path, "r") as opened_file:
+            with open(path) as opened_file:
                 template = opened_file.read()
             # Update the context with macros from the file.
             macro_ctx.update(
@@ -97,6 +97,27 @@ class JinjaTemplater(PythonTemplater):
             )
         return macro_ctx
 
+    def _extract_libraries_from_config(self, config):
+        library_path = config.get_section(
+            (self.templater_selector, self.name, "library_path")
+        )
+        if not library_path:
+            return {}
+
+        libraries = {}
+        for file_name in os.listdir(library_path):
+            file_path = os.path.join(library_path, file_name)
+            if not os.path.isfile(file_path) or not file_name.endswith(".py"):
+                continue
+
+            module_name = os.path.splitext(file_name)[0]
+            spec = importlib.util.spec_from_file_location(module_name, file_path)
+            lib = importlib.util.module_from_spec(spec)
+            spec.loader.exec_module(lib)
+            libraries[module_name] = lib
+
+        return libraries
+
     @staticmethod
     def _generate_dbt_builtins():
         """Generate the dbt builtins which are injected in the context."""
@@ -135,17 +156,10 @@ class JinjaTemplater(PythonTemplater):
             line_no = tree.lineno
             line = raw.split("\n")[line_no - 1]
             pos = line.index(tree.name) + 1
-            # Generate the charpos. +1 is for the newline characters themselves
-            charpos = (
-                sum(len(raw_line) + 1 for raw_line in raw.split("\n")[: line_no - 1])
-                + pos
-            )
-            # NB: The positions returned here will be *inconsistent* with those
-            # from the linter at the moment, because these are references to the
-            # structure of the file *before* templating.
             yield SQLTemplaterError(
-                "Undefined jinja template variable: {0!r}".format(tree.name),
-                pos=FilePositionMarker(None, line_no, pos, charpos),
+                f"Undefined jinja template variable: {tree.name!r}",
+                line_no=line_no,
+                line_pos=pos,
             )
 
     @staticmethod
@@ -160,7 +174,7 @@ class JinjaTemplater(PythonTemplater):
         )
 
     def process(
-        self, *, in_str: str, fname: Optional[str] = None, config=None
+        self, *, in_str: str, fname: str, config=None, formatter=None
     ) -> Tuple[Optional[TemplatedFile], list]:
         """Process a string and return the new string.
 
@@ -178,6 +192,7 @@ class JinjaTemplater(PythonTemplater):
                 mostly for loading config files at runtime.
             config (:obj:`FluffConfig`): A specific config to use for this
                 templating operation. Only necessary for some templaters.
+            formatter (:obj:`CallbackFormatter`): Optional object for output.
 
         """
         if not config:
@@ -201,18 +216,23 @@ class JinjaTemplater(PythonTemplater):
                 if name not in live_context:
                     live_context[name] = dbt_builtins[name]
 
-        # Load config macros
         env = self._get_jinja_env()
-        ctx = self._extract_macros_from_config(config=config, env=env, ctx=live_context)
+
         # Load macros from path (if applicable)
         macros_path = config.get_section(
             (self.templater_selector, self.name, "load_macros_from_path")
         )
         if macros_path:
-            ctx.update(
+            live_context.update(
                 self._extract_macros_from_path(macros_path, env=env, ctx=live_context)
             )
-        live_context.update(ctx)
+
+        # Load config macros, these will take precedence over macros from the path
+        live_context.update(
+            self._extract_macros_from_config(config=config, env=env, ctx=live_context)
+        )
+
+        live_context.update(self._extract_libraries_from_config(config=config))
 
         # Load the template, passing the global context.
         try:
@@ -225,17 +245,8 @@ class JinjaTemplater(PythonTemplater):
                 TemplatedFile(source_str=in_str, fname=fname),
                 [
                     SQLTemplaterError(
-                        "Failure to parse jinja template: {0}.".format(err),
-                        pos=FilePositionMarker(
-                            None,
-                            err.lineno,
-                            None,
-                            # Calculate the charpos for sorting.
-                            sum(
-                                len(line)
-                                for line in in_str.split("\n")[: err.lineno - 1]
-                            ),
-                        ),
+                        f"Failure to parse jinja template: {err}.",
+                        line_no=err.lineno,
                     )
                 ],
             )
@@ -250,9 +261,7 @@ class JinjaTemplater(PythonTemplater):
             undefined_variables = meta.find_undeclared_variables(syntax_tree)
         except Exception as err:
             # TODO: Add a url here so people can get more help.
-            raise SQLTemplaterError(
-                "Failure in identifying Jinja variables: {0}.".format(err)
-            )
+            raise SQLTemplaterError(f"Failure in identifying Jinja variables: {err}.")
 
         # Get rid of any that *are* actually defined.
         for val in live_context:
@@ -286,7 +295,7 @@ class JinjaTemplater(PythonTemplater):
             violations.append(
                 SQLTemplaterError(
                     (
-                        "Unrecoverable failure in Jinja templating: {0}. Have you configured "
+                        "Unrecoverable failure in Jinja templating: {}. Have you configured "
                         "your variables? https://docs.sqlfluff.com/en/latest/configuration.html"
                     ).format(err)
                 )
